#include <cstdio>
#include <sys/time.h> /* gettimeofday, setitimer */
#include <termio.h>   /* termio */

#include <sys/resource.h> /* PRIO_* */

#include <csignal>

#include <algorithm>
#include <list>

/* Captures video from a X window.
 * Requires the SHM extension to work on tolerable speeds.
 *
 * Copyright (C) 1992,2004 Bisqwit (http://iki.fi/bisqwit/)
 *
 * TODO: Create a standalone program without all parameters being
 *       hardcoded into the program.
 *
 * To compile: gcc -o xgetimage xgetimage.c -L/usr/X11R6/lib -lX11 -lXext -lXmu
 *
 */

#include "settings.hh"
#include "buffer.hh"
#include "avifun.hh"
#include "msg.hh"
#include "x.hh"

typedef unsigned long long UINT64;
static UINT64 getutime(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000LL) + tv.tv_usec;
}

static unsigned frames_total    = 0;
static UINT64 start_time;

static double GetVideoExcess()
{
    UINT64 time_now = getutime();
    double videopos = frames_total / (double)FPS;
    double want_pos = (time_now - start_time) / 1e6;
    return videopos - want_pos;
}
static void HandleImage(const unsigned char *vidbuf)
{
    unsigned frames_now = 0;
    
    // Set a high threshold for the first frame
    // to make it unlikely to discard frames
    double threshold = 0.333 / FPS;
    
    for(;;)
    {
        double excess = GetVideoExcess();
        if(excess > threshold) break; /* discard frame, we've had enough already */
        threshold = 0;

        SendFrame(vidbuf, width, height, !frames_total);
        ++frames_total;
        ++frames_now;
    }
    
    if(!frames_now)
        MessageDiscardFrame();
    else
        MessageCapturedFrame(frames_now);
}

static void InitAvi()
{
    OpenAVI();
    start_time   = getutime();
    frames_total = 0;
}

static void DoneAvi()
{
    CloseAVI();
}

void BufferEmptied(FILE *)
{
}


static termio back;
static void InitKB()
{
    struct termio term;
    ioctl(0, TCGETA, &term);
    ioctl(0, TCGETA, &back);
    term.c_lflag &= ~(ECHO | ICANON);
    term.c_cc[VMIN] = 0; 
    ioctl(0, TCSETA, &term);
}
static void DoneKB()
{
    ioctl(0, TCSETA, &back);
}

static void SetTimeVal(struct timeval& t, double value)
{
    t.tv_sec  = (unsigned) value;
    t.tv_usec = (unsigned) ((value - t.tv_sec) * 1e6);
}

static void CaptureFrame(int) { /* nothing to do */ }
static void InitializeTimers()
{
    struct itimerval tr;
    SetTimeVal(tr.it_interval, 1.0 / FPS);
    
    double next = GetVideoExcess() * DELAY_MULTIPLIER;

    // Set a minimum to ensure the timer won't be disabled
    if(next <= 1.4e-6) next = 1.4e-6;

    SetTimeVal(tr.it_value,    next);

    setpriority(PRIO_PROCESS, 0, -20);
    setitimer(ITIMER_REAL, &tr, NULL);
    
    /*
    *   FIXME: SEE ALSO "man 3p timer_create", "man 3p timer_getoverrun"
    *
    **/
}
static void EndTimers()
{
    struct itimerval tr;
    SetTimeVal(tr.it_value,    0);
    SetTimeVal(tr.it_interval, 0);
    setitimer(ITIMER_REAL, &tr, NULL);
    signal(SIGALRM, SIG_DFL);
}

int main(int argc, const char*const *argv)
{
    XInit();
    
    // Get 1 image to ensure there are no
    // first-initialization-delays waiting for us.
    GetImage();
    DeleteImage();
    
    InitKB();
    
    signal(SIGALRM, CaptureFrame);

    InitAvi();
    
    InitializeTimers();
    
    while(!BufferWaitInput())
    {
        const unsigned char *data = GetImage();
        if(data)
        {
            HandleImage(data);
            MessageFrameCounter(frames_total, GetBufferStatus());
        }
        DeleteImage();
        
        InitializeTimers();
    }
    
    EndTimers();
    
    DoneKB();
    DoneAvi();
    
    MessageEnd();
    
    XDone();
    
    return 0;
}
